/* tg- a command line tagger for mp4 and related files				(c) Richard 2003
 * need more error checking
 */ 

#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/utime.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <process.h>
#include <io.h>

#include "tg.h"
#include "atomparser.h"

#define DEBUG 1
#define TAGSIZE 750

void store4int(unsigned int n, char *data) {
  data[0] = (n >> 24) & 0xff;
  data[1] = (n >> 16) & 0xff;
  data[2] = (n >>  8) & 0xff;
  data[3] = (n >>  0) & 0xff;
}

void dprintf (const char *fmt, ...)
{
	va_list ap;

   if(DEBUG) {	
      va_start (ap, fmt);
      vprintf (fmt, ap);
      va_end (ap);
   }
}

	
int CreateUDTA(Song *tag, char *fulltag, int len) {

	static char udta[] = {0,0,0,0,'u','d','t','a'};
	static char meta[] = {0,0,0,0,'m','e','t','a',0,0,0,0};
	static char hdlr[] = {0,0,0,0,'h','d','l','r',0,0,0,0,0,0,0,0,'m','d','i','r',
		'a','p','p','l',0,0,0,0,0,0,0,0,0,0};
	static char ilst[] = {0,0,0,0,'i','l','s','t'};		
	static char cnam[] = {0,0,0,0,'','n','a','m',0,0,0,0,'d','a','t','a',0,0,0,1,0,0,0,0};
	static char cart[] = {0,0,0,0,'','A','R','T',0,0,0,0,'d','a','t','a',0,0,0,1,0,0,0,0};
	static char calb[] = {0,0,0,0,'','a','l','b',0,0,0,0,'d','a','t','a',0,0,0,1,0,0,0,0};		
	static char gnre[] = {0,0,0,0,'g','n','r','e',0,0,0,0,'d','a','t','a',0,0,0,0,0,0,0,0,0,0};
	static char trkn[] = {0,0,0,0,'t','r','k','n',0,0,0,0,'d','a','t','a',0,0,0,0,0,0,0,0,0,0,
		0,0,0,0,0,0};
	static char disk[] = {0,0,0,0,'d','i','s','k',0,0,0,0,'d','a','t','a',0,0,0,0,0,0,0,0,0,0,
		0,1,0,1};
	static char cday[] = {0,0,0,0,'','d','a','y',0,0,0,0,'d','a','t','a',0,0,0,1,0,0,0,0};
	static char cpil[] = {0,0,0,0,'c','p','i','l',0,0,0,0,'d','a','t','a',0,0,0,21,0,0,0,0,0};	
	static char tmpo[] = {0,0,0,0,'t','m','p','o',0,0,0,0,'d','a','t','a',0,0,0,21,0,0,0,0,0,0};		
	static char ctoo[] = {0,0,0,0,'','t','o','o',0,0,0,0,'d','a','t','a',0,0,0,1,0,0,0,0};
	static char cto2[] = {"iTunes v4.0, QuickTime 6.3"};
	static char ccmt[] = {0,0,0,0,'','c','m','t',0,0,0,0,'d','a','t','a',0,0,0,1,0,0,0,0};
	int pos;

// to do: make sure pos never exceeds sizeof(fulltag)

// append each atom to fulltag
// pos serves as a running total of the size of the atom 
	pos = 0;  
	memcpy(fulltag, udta, sizeof(udta));
	pos += sizeof(udta);
	memcpy(fulltag+pos, meta, sizeof(meta));
	pos += sizeof(meta);
	memcpy(fulltag+pos, hdlr, sizeof(hdlr));
	pos += sizeof(hdlr);
	memcpy(fulltag+pos, ilst, sizeof(ilst));
	pos += sizeof(ilst);
	
	// song title (string)
	if(tag->title != 0) {
   	store4int(sizeof(cnam)+strlen(tag->title), cnam);
   	store4int(sizeof(cnam)+strlen(tag->title)-8, cnam+8);
   	memcpy(fulltag+pos, cnam, sizeof(cnam));
   	pos += sizeof(cnam);
   	memcpy(fulltag+pos, tag->title, strlen(tag->title));
   	pos += strlen(tag->title);
	}
	
	// artist name (string)
	if(tag->artist != 0) {
   	store4int(sizeof(cart)+strlen(tag->artist), cart);
   	store4int(sizeof(cart)+strlen(tag->artist)-8, cart+8);
   	memcpy(fulltag+pos, cart, sizeof(cart));
   	pos += sizeof(cart);
   	memcpy(fulltag+pos, tag->artist, strlen(tag->artist));
   	pos += strlen(tag->artist);
	}
	
	// album name (string)
	if(tag->album != 0) {
   	store4int(sizeof(calb)+strlen(tag->album), calb);
   	store4int(sizeof(calb)+strlen(tag->album)-8, calb+8);
   	memcpy(fulltag+pos, calb, sizeof(calb));
   	pos += sizeof(calb);
   	memcpy(fulltag+pos, tag->album, strlen(tag->album));
   	pos += strlen(tag->album);
	}
	
	// genre number (id3 + 1)
	if(tag->genrenum != 0) {
   	store4int(sizeof(gnre), gnre);
   	store4int(sizeof(gnre)-8, gnre+8);
   	gnre[sizeof(gnre)-1] = tag->genrenum;
   	memcpy(fulltag+pos, gnre, sizeof(gnre));
   	pos += sizeof(gnre);
	}
	
	// track number & total tracks
	if(tag->track != 0) {
   	store4int(sizeof(trkn), trkn);
   	store4int(sizeof(trkn)-8, trkn+8);	
   	trkn[sizeof(trkn)-5] = tag->track;
   	trkn[sizeof(trkn)-3] = tag->track;
   	memcpy(fulltag+pos, trkn, sizeof(trkn));
   	pos += sizeof(trkn);
	}
	
	// disk number & total disks
	if(1) {
   	store4int(sizeof(disk), disk);
   	store4int(sizeof(disk)-8, disk+8);	
   	disk[sizeof(disk) - 3] = 1;
   	disk[sizeof(disk) - 1] = 1;
   	memcpy(fulltag+pos, disk, sizeof(disk));
   	pos += sizeof(disk);
	}
	
	// year (string)
	if(tag->year != 0) {	
   	store4int(sizeof(cday)+strlen(tag->year), cday);
   	store4int(sizeof(cday)+strlen(tag->year)-8, cday+8);
   	memcpy(fulltag+pos, cday, sizeof(cday));
   	pos += sizeof(cday);strlen(tag->year);
   	memcpy(fulltag+pos, tag->year, strlen(tag->year));
   	pos += strlen(tag->year);
	}
	
	// part of a compilation (0 or 1)
	if(1) {
   	store4int(sizeof(cpil), cpil);
   	store4int(sizeof(cpil)-8, cpil+8);	
   	memcpy(fulltag+pos, cpil, sizeof(cpil));
   	pos += sizeof(cpil);
	}
	
	// tempo in BPM
	if(1) {
   	store4int(sizeof(tmpo), tmpo);
   	store4int(sizeof(tmpo)-8, tmpo+8);	
   	memcpy(fulltag+pos, tmpo, sizeof(tmpo));
   	pos += sizeof(tmpo);
	}   	

	// tool used to create file (string)
	if(1) {
   	store4int(sizeof(ctoo)+sizeof(cto2)-1, ctoo);
   	store4int(sizeof(ctoo)+sizeof(cto2)-9, ctoo+8);	
   	memcpy(fulltag+pos, ctoo, sizeof(ctoo));
   	pos += sizeof(ctoo);
   	memcpy(fulltag+pos, cto2, sizeof(cto2)-1);
   	pos += sizeof(cto2)-1;
	}

	// comment (string)
	if(tag->comment != 0) {
   	store4int(sizeof(ccmt)+strlen(tag->comment), ccmt);
   	store4int(sizeof(ccmt)+strlen(tag->comment)-8, ccmt+8);
   	memcpy(fulltag+pos, ccmt, sizeof(ccmt));
   	pos += sizeof(ccmt);
   	memcpy(fulltag+pos, tag->comment, strlen(tag->comment));
   	pos += strlen(tag->comment);
	}

// done with adding tags
// now clean up the atomsizes and perhaps add a free atom
	
	store4int(pos-20-sizeof(hdlr), fulltag+20+sizeof(hdlr)); // size of ilst
	
	if(pos < (len + 8) ) {
		store4int((len-pos), fulltag+pos);
		memcpy(fulltag+pos+4, "free", 4);
		pos = len;
	}
	
	store4int(pos, fulltag);		// size of entire atom at beginning
	store4int(pos-8, fulltag+8);	// size of meta at meta position
	store4int(sizeof(hdlr), fulltag+20);  // size of hdlr at it position

	return pos;
} // createudta

// write iTunes m4a heading to beginning of file
// take the moov atom starting at byte 32 and copy it to end
// change original moov atom to free atom
// change atomsize of new moov atom to accomodate new tags
int MOOVtoEnd(FILE *f, int len, int moovpos) {
	char AtomHeader[10], *moov;
	int atomsize;
	char m4ahdr[] = {0,0,0,32,'f','t','y','p','m','4','a',32,0,0,0,0,
		'M','4','A',32,'m','p','4','i','s','o','m',0,0,0,0};

   if(moovpos == 0)
   	return 0;

	fseek(f, moovpos, SEEK_SET);

   fread(AtomHeader, 1, 8, f);
   atomsize = OneByte(AtomHeader);
//   memcpy(atomname, AtomHeader+4, 4);
//   atomname[4] = 0;

   // check whether this is necessary
   if(moovpos == 32) {
	   fseek(f, 0, SEEK_SET);         // move to beginning
   	fwrite(m4ahdr, 1, sizeof(m4ahdr), f);           // write a compliant header
   }
   
   fseek(f, moovpos, SEEK_SET);  	// move to start of moov atom
		if( (moov = (char *)malloc(atomsize)) == NULL) {
			printf("MOOVtoEND: memory allocation failed\n");
   		exit(1);
   	}
	fread(moov, 1, atomsize, f);   	// read entire moov atom
	   
   fseek(f, moovpos+4, SEEK_SET); // back to start of atom
   fwrite("free", 1, 4, f);       // change old moov atom to free
   
   fseek(f, 0, SEEK_END);         	// moov atom to end of file
   store4int(atomsize+len, moov);	// bump the len of the moov
   fwrite(moov, 1, atomsize, f);   // write the new moov atom
   
   free(moov);
   return 1;
} //moovtoend

void parsepath(char *path, Song *tag) {
	char dir[_MAX_DIR], name[_MAX_FNAME];
	int r;

	_splitpath(path, NULL, dir, name, NULL);

	tag->track = atoi(name);
	strcpy(tag->title, name+5);
	
	r = strlen(dir);
	dir[r-1] = 0;
	--r;
	while(--r)
		if(dir[r] == '/' || dir[r] == '\\') {
			strcpy(tag->album, dir+r+1);
			dir[r] = 0;
			break;
		}
		
	while(--r)
		if(dir[r] == '/' || dir[r] == '\\') {
			strcpy(tag->artist, dir+r+1);
			break;
		}


/*----------------------------------------
printf("parse:\n");
printf("artist     = %s\n", tag->artist);
printf("album      = %s\n", tag->album);
printf("title      = %s\n", tag->title);
printf("track      = %d\n", tag->track);
------------------------------------------*/
} // parsepath

void usage(void) {
printf(" 
Tagger for mp4's - uses iTunes format.  Tested only with QT created mp4's. 
** Use at your own risk **                    (c) Richard 2003
Version 0.33  2003-08-10

Usage: tg <filename> Options
Wildcards are supported in filename [more or less]

Options:
[none]      print existing tags
--debug     print existing tags and atom info
--remove    remove existing tags
--auto      parses filepath - must be in form artist\\album\\nn - title.ext
--artist  <value>
--album   <value>
--title   <value>
--track   <value>
--genre   <value> [id3v1 tags only]
--year    <value>
--comment <value>
 
use mp4creator.exe to optimize mp4 
thanks to www.getid3.org and tag for inspiration");    	

} // usage

int tagger(int argc, char *pfile, char *argv[]) {
	Song tag;
	FileInfo fi;
	char infile[MAX_PATH+1], outfile[MAX_PATH+1], fulltag[1024];
// char str[1024];
	char *buf;
	int end, changetags, i, moovpos, len, wrote;
	int autoo, back, debug, removetags;
	FILE *inf, *outf;
	struct stat tim;
	struct utimbuf ut;

	strncpy((char *)&tag, "", sizeof(tag));
	strncpy((char *)&fi, "", sizeof(fi));
	
	_fullpath(infile, pfile, 1023);
	sprintf(outfile, "%s.bak", infile);
	stat(infile, &tim);	// remember file times
	ut.actime = tim.st_atime;
	ut.modtime = tim.st_mtime;

	if( (inf  = fopen(infile, "rb")) == NULL) {
		printf("File not found: %s\n", infile);
		return 0;
	}
	
	fseek(inf, 0, SEEK_END);
	end = ftell(inf);
	fseek(inf, 0, SEEK_SET);

// read the existing tags
   ParseAtom(inf, 0, end, &fi);
	memcpy(&tag, &(fi.tag), sizeof(Song));
	
// if auto, read tag fields from pathname.
// cmd line tags override pathname, which is why this is before general cmd line read
    changetags = 0;
    autoo = 0;
    for(i = 1; i < argc; i++) {
        if(strcmp(argv[i], "--auto") == 0) {
	        autoo = 1;
           changetags = 2;
            parsepath(infile, &tag);
            break;
        }
    }

// read params from command line
    back   = 0;
    removetags = 0;
    debug  = 0;
    for(i = 1; i < argc; i++) {
	    if(strcmp(argv[i], "--artist") == 0) {
				strncpy(tag.artist, argv[++i], sizeof(tag.artist)-1);
            changetags = 1;
        }
        else if(strcmp(argv[i], "--album") == 0){
				strncpy(tag.album, argv[++i], sizeof(tag.album)-1);
            changetags = 1;
        }
        else if(strcmp(argv[i], "--title") == 0) {
				strncpy(tag.title, argv[++i], sizeof(tag.title)-1);
            changetags = 1;
        }
        else if(strcmp(argv[i], "--comment") == 0) {
				strncpy(tag.comment, argv[++i], sizeof(tag.comment)-1);
            changetags = 1;
        }        
        else if(strcmp(argv[i], "--track") == 0) {
	        tag.track = atoi(argv[++i]);
	        changetags = 1;
	     }
        else if(strcmp(argv[i], "--genre") == 0) {
				strncpy(tag.genrename, argv[++i],sizeof(tag.genrename)-1);
				tag.genrenum = genrecode(argv[i]);
				if(tag.genrenum == 255) {
					printf("Warning: unrecognized genre %s\n", argv[i]);
				}
            changetags = 1;
        }
        else if(strcmp(argv[i], "--year") == 0) {
				strncpy(tag.year, argv[++i], sizeof(tag.year)-1);		
            changetags = 1;
        }
        else if(strcmp(argv[i], "--back") == 0) {
	        back = 1;
	     }
        else if(strcmp(argv[i], "--debug") == 0) {
	        debug = 1;
	     }
        else if(strcmp(argv[i], "--remove") == 0) {
	        removetags = 1;
	        changetags = 1;
	     }
	     else if( strcmp(argv[i], "--auto") == 0) {
		     autoo = 1;
		  }
	     else {
            printf("unknown field: %s\n", argv[i]);
            FreeFileInfo(&fi);
            return 0;
        }  // if chain
    }  // for

/*----------------------------------------------
printf("auto       = %d\n", autoo);
printf("artist     = %s\n", tag.artist);
printf("album      = %s\n", tag.album);
printf("title      = %s\n", tag.title);
printf("track      = %d\n", tag.track);
printf("genre      = %s (%d)\n", tag.genrename, tag.genrenum);
printf("year       = %s\n", tag.year);
printf("back       = %d\n", back);
printf("debug      = %d\n", debug);
printf("remove     = %d\n", removetags);
printf("changetags = %d\n", changetags);
------------------------------------------------------*/

// if we  are not changing tags, just report tag info and return
	if(changetags == 0) {
   	if(debug) {
      	for(i = 0; i < fi.entries; i++)
	      	PrintAtom(fi.atoms[i]);
      } // if debug
      printf("\n%s\n", infile);
		if(fi.udtapos != 0)
            PrintTag(&fi.tag);
        else
            printf("no tags\n");
        FreeFileInfo(&fi);
        return 1;
    } // no change

// we are changing something, so create an outfile
	if( (outf = fopen(outfile, "w+b")) == NULL) {
		printf("Can't open %s\n", outfile);
		FreeFileInfo(&fi);
		return 0;
	}

//copy inf to outfile
	if( (buf = (char *)malloc(end)) == NULL) {
		printf("tagger: memory allocation failure\n");
		exit(1);
	}
	fseek(inf, 0, SEEK_SET);
	fread(buf, 1, end, inf);
	fwrite(buf, 1, end, outf);
	free(buf);
	fclose(inf);
	
//if remove, do it
    if(removetags == 1) {
        if(fi.udtapos == 0)
            printf("no existing tags\n");    
        else {
				for(i = 0; i < 5; i++) {
   				fseek(outf, fi.udtapos+4, SEEK_SET);
               fwrite("free", 1, 4, outf);
               fclose(outf);
//   				printf("removed tag at %d from %s\n", fi.udtapos, outfile);
   				outf = fopen(outfile, "r+b");
					strncpy((char *)&fi, "", sizeof(fi));	
					ParseAtom(outf, 0, end, &fi);		
					if(fi.udtapos == 0)
	   				break;
				} // for loop
				if(back == 0) {
					utime(outfile, &ut); // change file times to match original
					remove(infile);
					fclose(outf);
					rename(outfile, infile);
				}
				printf("tags removed from %s\n", infile);
        } // else
        FreeFileInfo(&fi);
        return 1;
    } // if remove

// this section writes the new tag

	len = CreateUDTA(&tag, fulltag, 0);	

//is there an existing udta atom we can overwrite?
// if there's enough room in the atom, overwrite
// if not, change to a free atom
   wrote = 0;
   if(fi.udtapos != 0) {
		if((len < fi.udtasize) && (wrote == 0)) {
      	CreateUDTA(&tag, fulltag, fi.udtasize);
         fseek(outf, fi.udtapos, SEEK_SET);
         fwrite(fulltag, 1, fi.udtasize, outf);
         wrote = 1;
       } // if
       else {
       	fseek(outf, fi.udtapos+4, SEEK_SET);
         fwrite("free", 1, 4, outf);
       }
   } // there's a udta

// if we didn't overwite an existing udta, add to a moov      
	if(wrote == 0) {
		len = CreateUDTA(&tag, fulltag, TAGSIZE);	
   	moovpos = fi.moovpos;
   	MOOVtoEnd(outf, len, moovpos);
   	fwrite(fulltag, 1, len, outf);
   }

// done.  close the file
// then re-open the file to check the tags
	fclose(outf);
	fclose(inf);
	
// unless we want a backup, rename temp file to original name and delete original  	
	if(back == 0) {
		remove(infile);
		rename(outfile, infile);
		strcpy(outfile, infile);
	}

// if we used moovtoend, there is a large free atom.  use mpcreator to eliminate it
/*-----------------------------------------
	if(wrote == 0) {
		sprintf(str, "\"%s\"", outfile);
		printf("about to apply mp4creator -optimize to %s\n", str);
		spawnlp(P_WAIT, "mp4creator60.exe", "mp4creator60.exe", 
			"-optimize", str, NULL); 
	}
--------------------------------------*/

// reopen the resulting file to read its tags
	strncpy((char *)&fi, "", sizeof(fi));	
  	outf = fopen(outfile, "rb");
	fseek(outf, 0, SEEK_END);
	end = ftell(outf);  
	ParseAtom(outf, 0, end, &fi);
	printf("\n%s\n", outfile);
   PrintTag(&fi.tag);

// change file times to match original
  	utime(outfile, &ut); 
			
	FreeFileInfo(&fi);
   return 1;	
}	// tagger

int pstrcmp(const void *p1, const void *p2) {
	return strcmp(*(char * const *)p1, *(char * const *)p2);
}


int main(int argc, char *argv[]) {
	char s[MAX_PATH+1], **filelist;
	char drive[_MAX_DRIVE+1], dir[_MAX_DIR+1], name[_MAX_FNAME+1], ext[_MAX_EXT+1];
	struct _finddata_t fd;
	int fhand, i, j;
		
	if(argc == 1) {
		usage();
		exit(1);
	}
	
// rudimentary wildcard processing	
	_splitpath(argv[1], drive, dir, name, ext);
	if(*name == 0) {
		sprintf(s, "%s%s*", drive, dir);
	}
	else {
		fhand = _findfirst(argv[1], &fd);
		if(fd.attrib & _A_SUBDIR && strchr(name, '*') == 0) {
				strcat(dir, name);
				strcat(dir, "\\");
				sprintf(s, "%s%s*", drive, dir);
		}
		else {
			strncpy(s, argv[1], MAX_PATH);		
		}
		_findclose(fhand);
	}

// use findfirst/findnext to create filelist - a list of all files
// matching our command line pattern
//	it should be legal to pass NULL to first use of realloc
// filelist = (char **)malloc(2); if not
	filelist = NULL;  
	i = 0;
	
	if ( (fhand = _findfirst(s, &fd)) != -1) {
		do {
			if((fd.attrib & _A_SUBDIR) == 0) {
         	if( (filelist = (char **)realloc(filelist, (i+1)*sizeof(int))) == NULL) {
	         	printf("main: memory allocation failure\n");
					exit(1);
				}
				sprintf(s, "%s%s%s", drive, dir, fd.name);         	
				if( (filelist[i] = (char *)malloc(strlen(s)+1)) == NULL) {
					printf("main: memory allocation failure\n");
					exit(1);
				}
         	strcpy(filelist[i], s);
         	i++;
			}
		} while(_findnext(fhand, &fd) != -1);
		_findclose(fhand);
	}

	qsort(filelist, i-1, sizeof(char *), pstrcmp);
	
// tag each file in the filelist
	for(j = 0; j < i; j++)
		tagger(argc-1, filelist[j], argv+1);

// free the space we allocated for filelist
// not necessary, but makes me feel better
	for(j = 0; j < i; j++)
		free(filelist[j]);
	free(filelist);
}

